import json
import weakref

import pika

from melange.messaging import MessagingDriver, Message


class RabbitMQDriver(MessagingDriver):
    def __init__(self, **kwargs):
        super().__init__()
        connection_parameters = pika.ConnectionParameters(**kwargs)
        self.connection = pika.BlockingConnection(connection_parameters)
        self.channel = self.connection.channel()
        self.channel.basic_qos(prefetch_count=1)

    def retrieve_messages(self, queue):
        method_frame, properties, body = self.channel.basic_get(queue=queue, no_ack=False)

        if not method_frame:
            return []

        message = Message(message_id=str(method_frame.delivery_tag),
                          content=json.loads(body.decode('utf-8')),
                          metadata=properties)
        return [message]

    def publish(self, content, topic):
        result = self.channel.basic_publish(exchange=topic,
                                            routing_key='',
                                            body=content,
                                            properties=pika.BasicProperties(
                                                delivery_mode=2,  # make message persistent
                                            ))

        return result

    def declare_topic(self, topic_name):
        self.channel.exchange_declare(exchange=topic_name,
                                      exchange_type='fanout',
                                      durable=True)

        return topic_name

    def acknowledge(self, message):
        self.channel.basic_ack(delivery_tag=int(message.message_id))

    def declare_queue(self, queue_name, topic_to_bind=None, dead_letter_queue_name=None):
        queue = self.channel.queue_declare(queue=queue_name, durable=True)
        if topic_to_bind:
            self.channel.queue_bind(exchange=topic_to_bind, queue=queue_name)

        return queue.method.queue, dead_letter_queue_name

    def close_connection(self):
        self.connection.close()

    def delete_queue(self, queue):
        self.channel.queue_delete(queue=queue)

    def delete_topic(self, topic):
        self.channel.exchange_delete(exchange=topic)



import weakref


class Message:
    def __init__(self, message_id, content, metadata):
        self.message_id = message_id
        self.content = content
        self.metadata = metadata


class MessagingDriver:
    def __init__(self):
        self._finalizer = weakref.finalize(self, self.close_connection)

    def declare_topic(self, topic_name):
        """
        Declares a topic exchange with the name "topic name" and
        returns an object that represent the topic

        :param topic_name: The name of the topic to create
        :return: An object that represents a topic. The type of the object
        is only relevant inside the context of the driver, so what you
        return as a topic will be passed in next calls to the driver
        where a topic is required
        """
        raise NotImplementedError

    def declare_queue(self, queue_name, topic_to_bind=None, dead_letter_queue_name=None):
        """
        Declares a queue with the name "queue_name". Optionally, this
         queue may be binded to the topic "topic_to_bind" and associated
         to a dead_letter_queue "dead_letter_queue_name" where messages that
         were unable to deliver will be placed.

        :param queue_name: The name of the queue to create
        :param topic_to_bind: The topic object where you will bind your queue
        :param dead_letter_queue_name: The name of the dead letter queue to
        create and associate to the queue "queue_name"
        :return: A tuple, with the first element being the object queue
        created, and the second element is the dead letter queue object.
        The type of the queue object is only relevant inside the context of the driver, so what you
        return as a queue will be passed in next calls to the driver
        where a queue is required
        """
        raise NotImplementedError

    def retrieve_messages(self, queue):
        """
        Returns a list of messages (instances of Message type) that have
        been received from the queue.

        :param queue: queue to poll
        :return: a list of messages to process
        """
        raise NotImplementedError

    def publish(self, content, topic):
        """
        Publishes the content to the topic. The content must be a
        string (which is the json representation of an event)
        """
        raise NotImplementedError

    def acknowledge(self, message):
        """
        Acknowledges a message so that it won't be redelivered by
        the messaging infrastructure in the future
        """
        raise NotImplementedError

    def close_connection(self):
        """
        Override this function if you want to use some finalizer code
         to shutdown your driver in a clean way
        """
        pass

    def delete_queue(self, queue):
        """
        Deletes the queue
        """
        raise NotImplementedError

    def delete_topic(self, topic):
        """
        Deletes the topic
        """
        raise NotImplementedError



from datetime import datetime

from marshmallow import Schema, fields


class EventSchema(Schema):
    event_type_name = fields.Str()
    occurred_on = fields.DateTime()


class EventMessage:

    # A constant that subscriber can use in their "listens_to" events to
    # tell they are interested in all the events that happen on their topic
    ALL = 'ALL'

    event_type_name = 'Default'

    def __init__(self, occurred_on=datetime.now()):
        self.occurred_on = occurred_on
        self.event_type_name = self.event_type_name

    def get_occurred_on(self):
        return self.occurred_on

    def get_event_type_name(self):
        return self.event_type_name



from datetime import datetime


class DomainEvent:
    def __init__(self, occurred_on=datetime.now()):
        self.occurred_on = occurred_on

    def get_occurred_on(self):
        return self.occurred_on



import logging
import threading

from .domain_event import DomainEvent
from .domain_subscriber import DomainSubscriber
from melange.infrastructure import Singleton


@Singleton
class DomainEventBus:
    thread_local = threading.local()

    def __init__(self):
        self.thread_local.subscribers = []
        self.thread_local.publishing = False

    def publish(self, event):
        if self.thread_local.publishing:
            return

        if not isinstance(event, DomainEvent):
            logging.error('Invalid data passed. You must pass an event instance')
            return

        self.thread_local.publishing = True
        self._publish_synchronous(event)
        self.thread_local.publishing = False

    def reset(self):
        val = getattr(self.thread_local, 'publishing', None)
        if val is None:
            val = False
            self.thread_local.publishing = False

        if not val:
            self.thread_local.subscribers = []

    def subscribe(self, subscriber):
        if not isinstance(subscriber, DomainSubscriber):
            return False

        if subscriber not in self.thread_local.subscribers:
            self.thread_local.subscribers.append(subscriber)

    def _publish_synchronous(self, event):
        subscribers = self._get_subscribers(event)

        for subscr in subscribers:
            try:
                subscr.process(event)
            except Exception as e:
                logging.error(e)

    def _get_subscribers(self, event):
        return [subscriber for subscriber in self.thread_local.subscribers if subscriber.accepts(event)]

    def get_publishing(self):
        val = getattr(self.thread_local, 'publishing', None)
        if val is None:
            self.thread_local.publishing = False

        return self.thread_local.publishing



""" Subscriber super-class """
from .domain_event_bus import DomainEvent


class DomainSubscriber:
    def process(self, event):
        """
        Called by the domain_event_bus.

        :param event: The event object
        :type event: EventMessage or subclass of event

        This method implements the logic for processing the event. This method should not block for
        long time as that will affect the performance of the domain_event_bus.
        """
        pass

    def listens_to(self):
        return []

    def accepts(self, event):
        return isinstance(event, DomainEvent) \
               and (self.listens_to() == [] or any(isinstance(event, e) for e in self.listens_to()))



from base64 import b64decode
from urllib.request import urlopen

import OpenSSL
from OpenSSL.crypto import load_certificate, FILETYPE_PEM, verify

SNS_MESSAGE_TYPE_SUB_NOTIFICATION = "SubscriptionConfirmation"
SNS_MESSAGE_TYPE_NOTIFICATION = "Notification"
SNS_MESSAGE_TYPE_UNSUB_NOTIFICATION = "UnsubscribeConfirmation"


def canonical_message_builder(content, fmt):
    """ Builds the canonical message to be verified.

        Sorts the fields as a requirement from AWS

        Args:
            content (dict): Parsed body of the response
            fmt (list): List of the fields that need to go into the message
        Returns (str):
            canonical message
    """
    m = ""

    for field in sorted(fmt):
        try:
            m += field + "\n" + content[field] + "\n"
        except KeyError:
            # Build with what you have
            pass

    return str(m)


def verify_sns_notification(message):
    """ Takes a notification request from Amazon push service SNS and verifies the origin of the notification.
        This implementation uses OpenSSL Crypto, inspired by the implementation of
        Artur Rodrigues with M2Crypto: http://goo.gl/KAgPPc
        Args:
            message (dictionary):
                The parsed message content
        Returns (bool):
            True if he message passes the verification, False otherwise
    """

    canonical_sub_unsub_format = ["Message", "MessageId", "SubscribeURL", "Timestamp", "Token", "TopicArn", "Type"]
    canonical_notification_format = ["Message", "MessageId", "Subject", "Timestamp", "TopicArn", "Type"]

    content = message
    decoded_signature = b64decode(content["Signature"])

    # Depending on the message type, canonical message format varies: http://goo.gl/oSrJl8
    if message['Type'] == SNS_MESSAGE_TYPE_SUB_NOTIFICATION or message['Type'] == SNS_MESSAGE_TYPE_UNSUB_NOTIFICATION:

        canonical_message = canonical_message_builder(content, canonical_sub_unsub_format)

    elif message['Type'] == SNS_MESSAGE_TYPE_NOTIFICATION:

        canonical_message = canonical_message_builder(content, canonical_notification_format)

    else:
        raise ValueError("Message Type (%s) is not recognized" % message['Type'])

    # Load the certificate and extract the public key
    cert = load_certificate(FILETYPE_PEM, urlopen(content["SigningCertURL"]).read().decode('utf-8'))

    try:
        verify(cert, decoded_signature, str.encode(canonical_message), 'sha1')
    except OpenSSL.crypto.Error as e:
        return False

    return True



import os

CACHE_REDIS_HOST = os.environ.get('CACHE_REDIS_HOST')
CACHE_REDIS_PORT = os.environ.get('CACHE_REDIS_PORT')
CACHE_REDIS_DB = os.environ.get('CACHE_REDIS_DB')
CACHE_REDIS_PASSWORD = os.environ.get('CACHE_REDIS_PASSWORD')
CACHE_NAMESPACE = os.environ.get('CACHE_NAMESPACE')



from atexit import register
from threading import Thread, Lock
from time import time

from .exchange_message_consumer import ExchangeMessageConsumer


class ThreadedExchangeMessageConsumer(Thread, ExchangeMessageConsumer):
    def __init__(self, event_queue_name, topic_to_subscribe=None, driver=None):

        ExchangeMessageConsumer.__init__(self, event_queue_name, topic_to_subscribe, driver=driver)
        Thread.__init__(self)

        register(self.shutdown)

        self.keep_running = True
        self.stop_time = 0
        self.shutdown_lock = Lock()

    def run(self):

        while not self._thread_should_end():
            self.consume_event()

    def _thread_should_end(self):
        return 0 < self.stop_time < time()

    def shutdown(self):
        """
        Stops the event bus. The event bus will stop all its executor threads.
        It will try to flush out already queued events by calling the subscribers
        of the events. This flush wait time is 2 seconds.
        """

        with self.shutdown_lock:
            if not self.keep_running:
                return
            self.keep_running = False
        self.stop_time = time() + 2

        self.join()



import json

import boto3

from melange.messaging import MessagingDriver, Message


class AWSDriver(MessagingDriver):
    def __init__(self):
        super().__init__()

    def declare_topic(self, topic_name):
        sns = boto3.resource('sns')
        topic = sns.create_topic(Name=topic_name)
        return topic

    def declare_queue(self, queue_name, topic_to_bind=None, dead_letter_queue_name=None):
        sqs_res = boto3.resource('sqs')

        queue = sqs_res.create_queue(QueueName=queue_name)

        if topic_to_bind:
            policy = {
                'Version': '2012-10-17',
                'Id': 'sqspolicy',
                'Statement': [
                    {
                        'Sid': 'First',
                        'Effect': 'Allow',
                        'Principal': '*',
                        'Resource': queue.attributes['QueueArn'],
                        'Action': 'sqs:SendMessage',
                        'Condition': {
                            'ArnEquals': {
                                'aws:SourceArn': topic_to_bind.arn
                            }
                        }
                    }
                ]
            }
            queue.set_attributes(Attributes={'Policy': json.dumps(policy)})
            topic_to_bind.subscribe(Protocol='sqs', Endpoint=queue.attributes['QueueArn'])

        dead_letter_queue = None
        if dead_letter_queue_name:
            dead_letter_queue = sqs_res.create_queue(QueueName=dead_letter_queue_name)

            redrive_policy = {
                'deadLetterTargetArn': dead_letter_queue.attributes['QueueArn'],
                'maxReceiveCount': '4'
            }

            queue.set_attributes(Attributes={'RedrivePolicy': json.dumps(redrive_policy)})

        return queue, dead_letter_queue

    def retrieve_messages(self, queue):
        messages = queue.receive_messages(MaxNumberOfMessages=1, VisibilityTimeout=100,
                                          WaitTimeSeconds=10, AttributeNames=['All'])

        return [Message(message.message_id, self._extract_message_content(message), message)
                for message in messages]

    def publish(self, content, topic):
        response = topic.publish(Message=content)

        if 'MessageId' not in response:
            raise ConnectionError('Could not send the event to the SNS TOPIC')

    def acknowledge(self, message):
        message.metadata.delete()

    def close_connection(self):
        pass

    def delete_queue(self, queue):
        queue.delete()

    def delete_topic(self, topic):
        topic.delete()

    def _extract_message_content(self, message):
        body = message.body
        message_content = json.loads(body)
        if 'Message' in message_content:
            content = json.loads(message_content['Message'])
        else:
            content = message_content

        return content



import json

from melange.messaging import EventSerializer


def parse_event_from_sns(sns_message):
    content = sns_message['Records'][0]['Sns']

    # TODO: Add message signature validation logic

    if 'Message' in content:
        content = content['Message']

    try:
        deserialized_dict = json.loads(content)
    except:
        raise Exception('The message content is not a valid json')

    return EventSerializer.instance().deserialize(deserialized_dict)



from redis_cache import SimpleCache, logging

from melange import settings
from .singleton import Singleton


@Singleton
class Cache:
    def __init__(self):
        self.cache = SimpleCache(expire=3600,
                                 host=settings.CACHE_REDIS_HOST,
                                 port=settings.CACHE_REDIS_PORT,
                                 db=settings.CACHE_REDIS_DB,
                                 password=settings.CACHE_REDIS_PASSWORD,
                                 namespace=settings.CACHE_NAMESPACE)

        if not self.cache.connection:
            logging.warning("Could not establish a connection with redis. Message deduplication won't work")

    def store(self, key, value, expire=None):
        if not self.cache.connection:
            return

        return self.cache.store(key, value, expire)

    def get(self, key):
        if not self.cache.connection:
            return

        return self.cache.get(key)

    def __contains__(self, key):
        return self.cache.connection and key in self.cache



from melange.drivers.aws import AWSDriver
from melange.drivers.rabbitmq_driver import RabbitMQDriver
from melange.messaging.driver_manager import DriverManager

DriverManager.instance().add_available_drivers(aws=AWSDriver,
                                               rabbitMQ=RabbitMQDriver)



from melange.infrastructure import Singleton
from .messaging_driver import MessagingDriver


@Singleton
class DriverManager:
    def __init__(self):
        self._driver = None
        self._drivers = {}

    def add_available_drivers(self, **kwargs):
        self._drivers.update(kwargs)

    def use_driver(self, driver=None, driver_name=None, **kwargs):
        if driver:
            if not isinstance(driver, MessagingDriver):
                raise Exception('Invalid driver supplied')

            self._driver = driver

        elif driver_name:
            driver = self._drivers.get(driver_name)
            if not driver:
                raise Exception('Invalid driver supplied')

            self._driver = driver(**kwargs)

        else:
            raise Exception('You need to either supply a driver or a driver_name!')

    def get_driver(self):
        if not self._driver:
            raise Exception('No driver is registered. Please call \'use_driver\' prior to getting it')

        return self._driver



from .driver_manager import *
from .event_message import *
from .event_serializer import *
from .exchange_listener import *
from .messaging_driver import *
from .threaded_exchange_message_consumer import *



from melange.infrastructure import Cache
from .event_message import EventMessage
from .utils import get_fully_qualified_name


class ExchangeListener:
    def process_event(self, event, **kwargs):
        if get_fully_qualified_name(self) + '.' + kwargs['message_id'] in Cache.instance():
            print('detected a duplicated message, ignoring')
            return

        self.process(event, **kwargs)

        Cache.instance().store(get_fully_qualified_name(self) + '.' + kwargs['message_id'],
                               get_fully_qualified_name(self) + '.' + kwargs['message_id'])

    def process(self, event, **kwargs):
        """
        Called by the domain_event_bus.

        :param event: The event object
        :type event: EventMessage or subclass of event

        This method implements the logic for processing the event. This method should not block for
        long time as that will affect the performance of the domain_event_bus.
        """
        pass

    def listens_to(self):
        # None = listens to every event
        return None

    def accepts(self, event_type_name):
        return not self.listens_to() or event_type_name in self.listens_to()



import json

from melange.infrastructure import Singleton
from .event_message import EventMessage


@Singleton
class EventSerializer:
    def __init__(self):
        self.event_serializer_map = {}

    def register(self, *serializers):
        for serializer in serializers:
            self.event_serializer_map[serializer.__qualname__] = serializer()

    def deserialize(self, event_dict):
        try:
            event_schema = event_dict['event_type_name'] + 'Schema'

            if event_schema not in self.event_serializer_map:
                raise ValueError("The event type {} doesn't have a registered serializer".format(event_dict['event_type_name']))

            schema = self.event_serializer_map[event_schema]
            data, errors = schema.load(event_dict)
            return data
        except ValueError:
            return event_dict

    def serialize(self, event):
        if isinstance(event, EventMessage):
            event_schema = event.event_type_name + 'Schema'

            if event_schema not in self.event_serializer_map:
                raise ValueError("The event type {} doesn't have a registered serializer".format(event.event_type_name))

            schema = self.event_serializer_map[event_schema]
            data, errors = schema.dumps(event)
            return data
        else:
            return json.dumps(event)





import logging

from .event_serializer import EventMessage
from .event_serializer import EventSerializer
from .driver_manager import DriverManager


class ExchangeMessagePublisher:
    def __init__(self, topic, driver=None):
        self._driver = driver or DriverManager.instance().get_driver()
        self._topic = self._driver.declare_topic(topic)

    def publish(self, event, event_type_name=None):
        if not isinstance(event, EventMessage) and not isinstance(event, dict):
            logging.error('Invalid data passed. You must pass an event instance or a dict')
            raise Exception('Invalid data passed. You must pass an event instance or a dict')

        if not event_type_name and (isinstance(event, dict) and 'event_type_name' not in event):
            raise Exception('When passing an event as a dict it has to include at least the event_type_name property')

        if event_type_name and isinstance(event, dict):
            event['event_type_name'] = event_type_name

        content = EventSerializer.instance().serialize(event)

        self._driver.publish(content, self._topic)

        return True



import logging

from melange.messaging import EventMessage
from melange.messaging import EventSerializer
from melange.messaging import ExchangeListener
from .driver_manager import DriverManager


class ExchangeMessageConsumer:
    def __init__(self, event_queue_name, topic_to_subscribe=None, dead_letter_queue_name=None, driver=None):
        self._exchange_listeners = []
        self._driver = driver or DriverManager.instance().get_driver()
        self._topic = self._driver.declare_topic(topic_to_subscribe) if topic_to_subscribe else None
        self._event_queue, self._dead_letter_queue = self._driver.declare_queue(event_queue_name,
                                                                                self._topic,
                                                                                dead_letter_queue_name)

    def subscribe(self, exchange_listener):
        if not isinstance(exchange_listener, ExchangeListener):
            return False

        if exchange_listener not in self._exchange_listeners:
            self._exchange_listeners.append(exchange_listener)

    def unsubscribe(self, exchange_listener):
        if exchange_listener in self._exchange_listeners:
            self._exchange_listeners.remove(exchange_listener)

    def consume_event(self):
        messages = self._driver.retrieve_messages(self._event_queue)

        for message in messages:
            try:
                self._process_message(message)
            except Exception as e:
                logging.error(e)

    def _get_subscribers(self, event_type_name):
        return [listener for listener in self._exchange_listeners if listener.accepts(event_type_name)]

    def _process_message(self, message):
        if 'event_type_name' not in message.content:
            # The message will be ignored but not deleted from the
            # queue. Let the dead letter queue handle it
            return True

        event = EventSerializer.instance().deserialize(message.content)
        event_type_name = event.event_type_name if isinstance(event, EventMessage) else event['event_type_name']
        subscribers = self._get_subscribers(event_type_name)

        successful = 0
        for subscr in subscribers:
            try:
                subscr.process_event(event, message_id=message.message_id)
                successful += 1
            except Exception as e:
                logging.error(e)

        if successful == len(subscribers):
            self._driver.acknowledge(message)
